mutable struct InequalityConstraint{T,N,Cs} <: AbstractConstraint{T,N}
    id::Int64
    name::String

    constraints::Cs
    pid::Int64
    # childid::Int64

    sols0::SVector{N,T}
    sols1::SVector{N,T}
    solγ0::SVector{N,T}
    solγ1::SVector{N,T}
    

    function InequalityConstraint(data...; name::String="")
        bounddata = Tuple{Bound,Int64}[]
        for info in data
            if typeof(info[1]) <: Bound
                push!(bounddata, info)
            else
                for subinfo in info
                    push!(bounddata, subinfo)
                end
            end
        end

        T = getT(bounddata[1][1])

        pid = bounddata[1][2]
        childids = Int64[]
        constraints = Bound{T}[]
        N = 0
        for set in bounddata
            push!(constraints, set[1])
            @assert set[2] == pid
            N += 1 # getN(set[1])
        end
        constraints = Tuple(constraints)
        # Nc = length(constraints)

        sols0 = ones(T, N)
        sols1 = ones(T, N)
        solγ0 = ones(T, N)
        solγ1 = ones(T, N)

        new{T,N,typeof(constraints)}(getGlobalID(), name, constraints, pid, sols0, sols1, solγ0, solγ1)
    end
end


Base.length(::InequalityConstraint{T,N}) where {T,N} = N

function resetVars!(ineqc::InequalityConstraint{T,N}) where {T,N}
    ineqc.sols0 = @SVector ones(T, N)
    ineqc.sols1 = @SVector ones(T, N)
    ineqc.solγ0 = @SVector ones(T, N)
    ineqc.solγ1 = @SVector ones(T, N)

    return 
end

@inline function NtγTof!(mechanism, body::Body, ineqc::InequalityConstraint{T,N}) where {T,N}
    body.f -= ∂g∂pos(mechanism, ineqc, body)' * ineqc.solγ1
    for i=1:N
        body.f -= additionalforce(ineqc.constraints[i])
    end
    return
end

function g(mechanism, ineqc::InequalityConstraint{T,1}) where {T}
    g(ineqc.constraints[1], getbody(mechanism, ineqc.pid), mechanism.Δt)
end

@generated function g(mechanism, ineqc::InequalityConstraint{T,N}) where {T,N}
    vec = [:(g(ineqc.constraints[$i], getbody(mechanism, ineqc.pid), mechanism.Δt)) for i = 1:N]
    :(SVector{N,T}($(vec...)))
end

function gs(mechanism, ineqc::InequalityConstraint{T,1}) where {T}
    g(ineqc.constraints[1], getbody(mechanism, ineqc.pid), mechanism.Δt) - ineqc.sols1[1]
end

@generated function gs(mechanism, ineqc::InequalityConstraint{T,N}) where {T,N}
    vec = [:(g(ineqc.constraints[$i], getbody(mechanism, ineqc.pid), mechanism.Δt) - ineqc.sols1[$i]) for i = 1:N]
    :(SVector{N,T}($(vec...)))
end

function h(ineqc::InequalityConstraint)
    ineqc.sols1 .* ineqc.solγ1
end

function hμ(ineqc::InequalityConstraint{T}, μ) where T
    ineqc.sols1 .* ineqc.solγ1 .- μ
end


function schurf(mechanism, ineqc::InequalityConstraint{T,N}, body) where {T,N}
    val = @SVector zeros(T, 6)
    for i = 1:N
        val += schurf(ineqc, ineqc.constraints[i], i, body, mechanism.μ, mechanism.Δt)
    end
    return val
end

function schurD(ineqc::InequalityConstraint{T,N}, body, Δt) where {T,N}
    val = @SMatrix zeros(T, 6, 6)
    for i = 1:N
        val += schurD(ineqc, ineqc.constraints[i], i, body, Δt)
    end
    return val
end

@generated function ∂g∂pos(mechanism, ineqc::InequalityConstraint{T,N}, body) where {T,N}
    vec = [:(∂g∂pos(ineqc.constraints[$i])) for i = 1:N]
    :(vcat($(vec...)))
end

@generated function ∂g∂vel(mechanism, ineqc::InequalityConstraint{T,N}, body) where {T,N}
    vec = [:(∂g∂vel(ineqc.constraints[$i], mechanism.Δt)) for i = 1:N]
    :(vcat($(vec...)))
end

function calcFrictionForce!(mechanism, ineqc::InequalityConstraint{T,N}) where {T,N}
    for i = 1:N
        constraint = ineqc.constraints[i]
        if typeof(constraint) <: Friction
            calcFrictionForce!(mechanism, ineqc, constraint, i, getbody(mechanism, ineqc.pid))
        end
    end
end